#ifndef _CKPT_H
#define _CKPT_H

#include <setjmp.h>
#include <signal.h>
#include <sys/types.h>

#ifdef TIME_CKPT
#include <limits.h>
#include <sys/times.h>
#endif


/*
#define CKPTTEMPFILE	"/home/frank/CKPT"
#define CKPTDIR		"/home/frank/ckpt"
#define LOGFILE		"/home/frank/dtest/pout"
*/

#define CKPTSIGNAL	SIGUSR1
#define CKPTPORT	4666

static enum { MACHINE_SUN4,
	      MACHINE_SUN4C, 
	      MACHINE_SUN4M, 
	      MACHINE_I486, 
	      MACHINE_RS6000,
	    } mach_t;

struct _ckpt_info {
  caddr_t	heap_start;
  caddr_t	heap_end;
  u_long	heap_size;
  caddr_t	stack_top;
  caddr_t	stack_bottom;
  u_long	stack_size;
  int		mach;
};
typedef struct _ckpt_info ckpt_info;

#ifndef raise
#define raise(SIG) kill(getpid(), SIG)
#endif

/* from ckpt_main.c */
int	ckpt_startup(int, char **);
void	install_ckpt_handler(void);
void	ckpt_handler(int);
void	ckpt_restart(void);

/* from ckpt_util.c */
void	ckpt_error(int, char *, ...);
int	get_mach();
int	Read(int, void *, int);
int	Write(int, void *, int);
#ifdef TIME_CKPT
void    report_time(clock_t, char *);
#endif

/* from ckpt_dump.c */
void	dump_checkpoint();
void	dump_header(int, ckpt_info);
void	dump_signals(int);
void	dump_heap(int, ckpt_info);
void	dump_jmpbuf(int);
void	dump_stack(int, ckpt_info);

/* from ckpt_restore.c */
void	restore_checkpoint(char *);
void	restore_header(int, ckpt_info *);
void	restore_signals(int);
void	restore_heap(int, ckpt_info);
void	restore_jmpbuf(int, ckpt_info);
void	restore_stack(int, ckpt_info);

/* from ckpt_io.c */
int	open_checkpoint_for_write();
int	open_checkpoint_for_read(char *);
int	close_checkpoint(int);

/* from ckpt_fio.c */
void	filestate_init();
void	filestate_display();
void	filestate_save();
void	filestate_open();

/* from ckpt_sys.c */
void	get_ckpt_info(ckpt_info *);

#endif _CKPT_H
