/*
 * File: error.c
 *
 * ABSTRACT:
 *  Routines for error handling
 * 
 * (C) University of Amsterdam 
 *     Section Computational Science 
 *     Kruislaan 403, the Netherlands
 *
 * CREATION DATE: Unknown
 * AUTHOR       : Z.W. Hendrikse
 * 
 * MODIFICATION HISTORY:
 * 
 * Date        By    Description
 * ----------  ----  -----------------------------------
 * 30-08-1999  ZWH   Added comments
 */

#include <stdio.h>
#include <stdarg.h>
#include <errno.h>
#include <string.h>
#include "ckpt.h"
#include "dpvmError.h"

/**
 * Construct an error message.
 * See error.h for possible <err_type>'s. 
 */
extern void error(int err_type, char *fmt, ...)
{
  va_list args;
  char errbuf[256];

#ifndef DEBUG
  if (err_type & ERR_DEBUG)
      return;
#endif

  va_start(args, fmt);
  vsprintf(errbuf, fmt, args);
  va_end(args);

  /*
   * Print system error if needed
   */
  if (err_type & ERR_PERR) {
      pvmlogerror("voor");
      strcat(errbuf, ": ");
      strcat(errbuf, strerror(errno));
      strcat(errbuf, "\n");
      pvmlogerror("na");
  }

  pvmlogerror(errbuf);
}
