#include <pthread.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#include <stdlib.h>
#include <stdio.h>

#include <rpc/types.h>
#include <rpc/xdr.h>

#include <sys/time.h>
#include <sys/select.h>

#include <string.h>

#include "monlib.h"

char *montypes[] = {"SEND", "RECV"};

int
main(argc, argv)
	int argc;
	char **argv;
{
	int sock, on = 1;
	struct sockaddr_in sin;
	int e = sizeof sin, s, i;
	fd_set readfds, errorfds, rfds, efds;
	int maxfd, error, nconn;
	char *buf;
	XDR xdr;
	struct mon_data md;
	char *dummy;
	int port, n;

	buf = malloc(sizeof(struct mon_data));
	
	memset(&sin, 0, sizeof (struct sockaddr_in));

	sock = socket(AF_INET, SOCK_STREAM, 0);
	if (sock < 0) {
		perror("socket");
		return;
	}

	if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char *) &on,
	    sizeof on) < 0) {
		perror("setsockopt");
		return;
	}

	mon_getmonaddr(&dummy, &port);

	sin.sin_family = AF_INET;
#ifdef HAS_SINLEN
	sin.sin_len = sizeof(struct sockaddr_in);
#endif
	sin.sin_addr.s_addr = htonl(INADDR_ANY);
	sin.sin_port = port;
	bind(sock, (struct sockaddr *) &sin, sizeof sin);
	listen(sock, 1024);

	FD_ZERO(&readfds);
	FD_ZERO(&errorfds);

	FD_SET(sock, &readfds);
	FD_SET(sock, &errorfds);

	maxfd = sock;

	while (1) {
		rfds = readfds;
		efds = errorfds;
		error = select(maxfd + 1, &rfds, NULL, &efds, NULL);
		if (FD_ISSET(sock, &rfds)) {
			s = accept(sock, (struct sockaddr *) &sin, &e);
			if (s < 0)
				perror("accept");
			else {
				FD_SET(s, &readfds);
				FD_SET(s, &errorfds);
				nconn++;
				if (s > maxfd)
					maxfd = s;
			}
			FD_CLR(sock, &rfds);
		}
		for (i = 0; i <= maxfd; i++) {
			if (FD_ISSET(i, &efds)) {
				FD_CLR(i, &readfds);
				FD_CLR(i, &errorfds);
				if (i == maxfd)
					maxfd--;
				continue;
			}
			if (FD_ISSET(i, &rfds)) {
				n = read(i, buf, sizeof(struct mon_data));
				if (n <= 0) {
					FD_CLR(i, &readfds);
					FD_CLR(i, &errorfds);
					if (i == maxfd)
						maxfd--;
					continue;
				}
				xdrmem_create(&xdr, buf,
				    sizeof(struct mon_data), XDR_DECODE);
				xdr_mondata(&xdr, &md);
				xdr_destroy(&xdr);
				/*
				 * Do something useful with the data.
				 */
				printf("%s: %x -> %x type %d len %d start %ld.%ld end %ld.%ld\n",
					montypes[md.mon_montype],
					md.mon_src, md.mon_dst, md.mon_msgtype,
					md.mon_len, md.mon_stime.tv_sec,
					md.mon_stime.tv_nsec, md.mon_etime.tv_sec,
					md.mon_etime.tv_nsec);
			}
		}
	}
}
