#ifndef _MONLIB_H
#define _MONLIB_H

#include <pthread.h>

typedef struct semaphore {
     pthread_mutex_t lock;
     pthread_cond_t  cond;
     int count;
} *sem_t;

struct mon_data {
	int mon_montype;
	int mon_msgtype;
	int mon_src;
	int mon_dst;
	int mon_len;
	struct timespec mon_stime;
	struct timespec mon_etime;
	struct mon_data *mon_next;
};

#define MON_TYPE_SEND	0
#define MON_TYPE_RECV	1

struct mon_list {
	struct mon_data *head;
	struct mon_data *tail;
	sem_t avail;
	pthread_mutex_t lock;
	pthread_t thread;
	int flags;
};

#define MON_EXITING	0x0001

sem_t sem_create(int val);
int sem_init(sem_t *sem, int val);
void sem_wait(sem_t *s);
void sem_signal(sem_t *s);
int sem_testwait(sem_t *s);
int sem_destroy(sem_t *s);
struct mon_list *monlist_create(void);
void monlist_init(struct mon_list *mp);
void monlist_addtail(struct mon_list *mp, struct mon_data *dp);
void monlist_remhead(struct mon_list *mp, struct mon_data *dp);
int mon_get_data(struct mon_list *mp, struct mon_data *mdp);
void mon_add_data(struct mon_list *mp, int montype, int type, int src, int dst,
		  int len, struct timespec *stime, struct timespec *etime);
void monlist_destroy(struct mon_list *mp);
void monlist_exit(struct mon_list *mp);
int mon_getmonaddr(char **host, int *port);

int mon_createthread(struct mon_list *mlp);
void mon_terminatethread(struct mon_list *mlp);

int xdr_mondata(XDR *xdrs, struct mon_data *mdp);
int xdr_timespec(XDR *xdrs, struct timespec *tsp);

extern struct mon_list __pvm_ml;
#endif
