/*
 *	route.c
 *
 *	Routetables
 *
 */

#include <stdlib.h>

#include "global.h"
#include "fromlib.h"
#include "protoglarp.h"
#include "pvmalloc.h"
#include "listmac.h"
#include "bfunc.h"
#include "route.h"

struct route *routes = 0;

static char pvmtxt[512];		/* scratch for error log */

void route_init()
{
  if (!routes) {
    routes = TALLOC(1, struct route, "route");
    BZERO((char*)routes, sizeof(struct route));
    routes->r_link = routes;
    routes->r_rlink = routes;
  }
}

void route_new(int tid, int route)
{
  struct route *rp, *rp2;
  
  if (!(rp = TALLOC(1, struct route, "route"))) {
    pvmlogerror("route_new() can't get memory\n");
    pvmbailout(0);
  }
  BZERO((char*)rp, sizeof(struct route));
  rp->r_tid   = tid;
  rp->r_route = route;
  rp->r_flag  = RF_OK;
  
  for (rp2 = routes->r_link; rp2 != routes; rp2 = rp2->r_link)
      if (rp2->r_tid > tid)
	  break;
  LISTPUTBEFORE(rp2, rp, r_link, r_rlink);
}

void route_free(struct route *rp)
{
  if (rp->r_link && rp->r_rlink)
      LISTDELETE(rp, r_link, r_rlink);
  PVM_FREE(rp);
}

struct route *route_find(int tid)
{
  struct route *rp;

  if (routes == NULL)
    return (struct route *)0;

  if (tid) {
    rp = routes->r_link;
    while (rp != routes && rp->r_tid < tid)
	rp = rp->r_link;
    if (rp->r_tid == tid)
	return rp;
  }
  return (struct route *)0;
}

void route_dump()
{
  struct route *rp;

  pvmlogerror("route_dump()\n");
  pvmlogerror("tid\t route\t flag\n");
  for (rp = routes->r_link; rp != routes; rp = rp->r_link) {
    sprintf(pvmtxt, "%8x %8x %8x\n", rp->r_tid, rp->r_route, rp->r_flag);
    pvmlogerror(pvmtxt);
  }
  pvmlogerror("\n");
}

void route_update(int tid, int route)
{
  struct route *rp;

  sprintf(pvmtxt, "route_update(%x, %x)\n", tid, route);
  pvmlogerror(pvmtxt);

  if (rp = route_find(tid)) 
      rp->r_route = route;
  else
      route_new(tid, route);

  route_dump();
}
