
/*
 *         PVM version 3.3:  Parallel Virtual Machine System
 *               University of Tennessee, Knoxville TN.
 *           Oak Ridge National Laboratory, Oak Ridge TN.
 *                   Emory University, Atlanta GA.
 *      Authors:  A. L. Beguelin, J. J. Dongarra, G. A. Geist,
 *    W. C. Jiang, R. J. Manchek, B. K. Moore, and V. S. Sunderam
 *                   (C) 1992 All Rights Reserved
 *
 *                              NOTICE
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear in
 * supporting documentation.
 *
 * Neither the Institutions (Emory University, Oak Ridge National
 * Laboratory, and University of Tennessee) nor the Authors make any
 * representations about the suitability of this software for any
 * purpose.  This software is provided ``as is'' without express or
 * implied warranty.
 *
 * PVM version 3 was funded in part by the U.S. Department of Energy,
 * the National Science Foundation and the State of Tennessee.
 */

/*
 *	tdpro.h
 *
 *	Task-pvmd protocol
 *
$Log: tdpro.h,v $
Revision 1.8  2000/03/10 10:42:08  kamil
Merge PvmMoveRouteUpdate into PvmMoveCheckpoint.

Revision 1.7  2000/02/01 09:18:56  kamil
Define TC_REALLAST.

Revision 1.6  2000/01/14 09:58:06  kamil
Put TC_MOVED back before TC_LAST.

Revision 1.5  2000/01/11 09:57:01  kamil
Move TC_MOVED behind TC_LAST, add TC_EOC.

 * Revision 1.4  1999/06/18  09:17:01  frank
 * Deal with out-of-sequence fragments. Make the daemons do forwarding. Check
 * messages' destination at the last moment with a routing table check. Solve
 * a racecondition in t-d communication when migrating.
 *
Revision 1.3  1999/04/27 09:50:25  frank
* handle direct connections
* don't re-initialize sockets when restarting, and certainly do not
  throw the buffers away

Revision 1.2  1999/04/15 08:28:26  frank
Checkin of half-baked direct connection support, to set a mark for the
latest version exported to Paderborn.

 * Revision 1.1.1.1  1998/09/25  12:43:50  frank
 * First Dynamite import
 *
 * Revision 1.3  1994/06/03  20:38:31  manchek
 * version 3.3.0
 *
 * Revision 1.2  1993/10/04  20:30:50  manchek
 * added alternate TDSOCKNAME if SHAREDTMP is defined
 *
 * Revision 1.1  1993/08/30  23:26:52  manchek
 * Initial revision
 *
 */


/* protocol version */

#ifndef	TDPROTOCOL
#define	TDPROTOCOL	1319
#endif

/* message/frag hdr sizes */

#define	TDFRAGHDR	20
#define	TTMSGHDR	20
#define TTFRAGHDR	 8
#define	MAXHDR		((DDFRAGHDR > TDFRAGHDR ? DDFRAGHDR : TDFRAGHDR) \
					+ (DDMSGHDR > (TTMSGHDR + TTFRAGHDR) ? DDMSGHDR : (TTMSGHDR + TTFRAGHDR)))

#ifdef	SHAREDTMP
#define	TDSOCKNAME	"/home/%s/pvm3/pvmd.%d.%s"	/* file with task->pvmd sockaddr */
#else
#define	TDSOCKNAME	"/tmp/pvmd.%d"	/* file with task->pvmd sockaddr */
#endif

#define	TM_FIRST	0x80010001		/* first TM_ message */
#define	TM_CONNECT	(TM_FIRST+0)	/* task connects */
#define	TM_CONN2	(TM_FIRST+1)	/* 2nd round of connect */
#define	TM_EXIT		(TM_FIRST+2)	/* task disconnects */
#define	TM_ADDHOST	(TM_FIRST+3)	/* add hosts to machine */
#define	TM_DELHOST	(TM_FIRST+4)	/* delete hosts from machine */
#define	TM_CONFIG	(TM_FIRST+5)	/* get machine config */
#define	TM_MSTAT	(TM_FIRST+6)	/* get host status */
#define	TM_HALT		(TM_FIRST+7)	/* halt machine */
#define	TM_TICKLE	(TM_FIRST+8)	/* tickle pvmd */
#define	TM_SPAWN	(TM_FIRST+9)	/* spawn new tasks */
#define	TM_PSTAT	(TM_FIRST+10)	/* get task status */
#define	TM_SENDSIG	(TM_FIRST+11)	/* signal a task */
#define	TM_TASK		(TM_FIRST+12)	/* list tasks */
#define	TM_MCA		(TM_FIRST+13)	/* request multicast tid */
#define	TM_NOTIFY	(TM_FIRST+14)	/* request notification */
#define	TM_DB		(TM_FIRST+15)	/* name-server request */
#define	TM_SCHED	(TM_FIRST+16)	/* register task as scheduler */
#define	TM_TASKER	(TM_FIRST+17)	/* register task as task starter */
#define	TM_HOSTER	(TM_FIRST+18)	/* register task as pvmd starter */
#define	TM_HOSTSYNC	(TM_FIRST+19)	/* get clock delta to host */
#define	TM_SETOPT	(TM_FIRST+20)	/* set task parameters */
#ifdef DPVM
#define TM_MOVE		(TM_FIRST+21)	/* task migration */
#define TM_MIG		(TM_FIRST+22)	/* task migration */
#define	TM_LAST		(TM_FIRST+22)	/* last TM_ message */
#else  DPVM
#define	TM_LAST		(TM_FIRST+20)	/* last TM_ message */
#endif DPVM

#define	TMDB_INSERT	1
#define	TMDB_DELETE	2
#define	TMDB_LOOKUP	3

/*
* for TM_SETOPT
*/

#define	TS_OUTTID	1				/* stdout dst */
#define	TS_OUTCOD	2				/* stdout message tag */
#define	TS_TRCTID	3				/* trace dst */
#define	TS_TRCCOD	4				/* trace message tag */

#define	TC_FIRST	0x80030001		/* first TC_ message */
#define	TC_CONREQ	(TC_FIRST+0)	/* connection request */
#define	TC_CONACK	(TC_FIRST+1)	/* connection request ack */
#define	TC_TASKEXIT	(TC_FIRST+2)	/* task exited/doesn't exist */
#define	TC_NOOP		(TC_FIRST+3)	/* do nothing */
#define	TC_OUTPUT	(TC_FIRST+4)	/* child stdout data */
#define	TC_SETTMASK	(TC_FIRST+5)	/* set task trace mask */
#ifdef DPVM
#define TC_MOVED	(TC_FIRST+6)
#define	TC_LAST		(TC_FIRST+6)	/* last TC_ message */
     /* This define is deliberately after TC_LAST, so that
	pvmmctl() is not invoked when it arrives.  */
#define TC_EOC		(TC_FIRST+7)    /* end of connection marker */
#define TC_REALLAST     (TC_FIRST+7)
#else
#define	TC_LAST		(TC_FIRST+5)	/* last TC_ message */
#endif


#ifdef DPVM

/*
 *	for pvm_move
 */

#define PvmMoveCreateContext	1
#define PvmMoveRouteBroadcast	2
#define PvmMoveCheckpoint	3
#define PvmMoveRestart		4

#endif DPVM

/* task debug mask XXX should not be here */

#define TDMPACKET       1               /* packet tracing */
#define TDMMESSAGE      2               /* message tracing */
#define TDMSELECT       4               /* select and fd sets */
#define TDMROUTE        8               /* message route control */

